#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define RIGHT_CHILD(i)  (((i) * 2) + 2)
#define PARENT(i)       (((i) - 1) / 2)
#define INIT_SIZE   1
#define GROW_SIZE	32

/*
* Macro
*/

#ifndef TRUE
#define TRUE	(!FALSE)
#endif

#ifdef DEBUG
#define return_if_fail(expr) \
{ \
	if (!(expr)) { \
		fprintf(stderr, "%s:%d:%s(): expression `%s' failed\n", \
			__FILE__, \
			__LINE__, \
			__FUNCTION__, \
			#expr); \
		return; \
	} \
}

#define return_val_if_fail(expr, val) \
{ \
	if (!(expr)) { \
		fprintf(stderr, "%s:%d:%s(): expression `%s' failed\n", \
			__FILE__, \
			__LINE__, \
			__FUNCTION__, \
			#expr); \
		return (val); \
	} \
}

#define assert(expr) \
{ \
	if (!(expr)) { \
		fprintf(stderr, "%s:%d:%s(): Assertion (%s) failed\n", \
			__FILE__, \
			__LINE__, \
			__FUNCTION__, \
			#expr); \
		abort(); \
	} \
}

#else
#define return_if_fail(expr)
#define return_val_if_fail(expr, val)
#define assert(expr)
#endif

typedef int (*prioq_compare_t)(struct prioq_head *a, struct prioq_head *b);

struct prioq_head {
	unsigned int index;
};		

struct user_data {
	struct prioq_head head;
	int               data;
};		

struct prioq {
	struct prioq_head   **heap;
	size_t              size;
	size_t              used;
	prioq_compare_t     compare;
};	



struct prioq *
prioq_new(prioq_compare_t compare)
{
	struct prioq *queue;

	queue = (struct prioq*)malloc(sizeof(struct prioq));
	//return_val_if_fail(queue !=NULL, NULL);
	if (queue == NULL) {
			fprintf(stderr,"prioq_new: can't allocate queue");
			return NULL;
	}	
	
	memset(queue, 0, sizeof(*queue));

	queue->heap = (struct prioq_head**)malloc(sizeof(struct prioq_head *) *INIT_SIZE);

	if (queue->heap == NULL) {
			fprintf(stderr, "prioq_new: can't allocate heap\n");
			free(queue);
			return NULL;
	}

	memset(queue->heap, 0, sizeof(struct prioq_head *) * INIT_SIZE);

	queue->size    = INIT_SIZE;
	queue->used    = 0;
	queue->compare = ((compare != NULL) ? compare : compare);
	return queue;
}

#define NEW_SIZE(q)	(((q)->size + GROW_SIZE) * sizeof(struct prioq_head *))
#define OLD_SIZE(q)	((q)->size * sizeof(struct prioq_head *))

int 
prioq_put(struct prioq *queue, struct prioq_head *hp)
{
	int i;
	struct prioq_head** tmp;
		
	if (queue == NULL || hp == NULL) {
		fprintf(stderr, "prioq_put: NULL args\n");
		return 0; 
	}
	if (queue->size == queue->used) {
		tmp = (struct prioq_head**)realloc(queue->heap, NEW_SIZE(queue));
		        
		if (!tmp) {
			fprintf(stderr, "prioq_put: can't realloc heap\n");
			return -1;
		}
		queue->heap = tmp;
		
		memset(queue->heap + queue->size, 0,
		NEW_SIZE(queue) - OLD_SIZE(queue));
				
		queue->size += GROW_SIZE;
	}

	i = queue->used;
	
	while (i && queue->compare(queue->heap[PARENT(i)], hp)< 0) {
			queue->heap[i]        = queue->heap[PARENT(i)];
			queue->heap[i]->index = i;
			i = PARENT(i);
	}
	queue->heap[i]        = hp;
	queue->heap[i]->index = i;
	queue->used++;
	return 0;
}	


int
compare(struct prioq_head *a, struct prioq_head *b)
{
	struct user_data *ap = (struct user_data *) a;
	struct user_data *bp = (struct user_data *) b;

	if (ap->data > bp->data)
			return 1;
	else if (ap->data < bp->data)
			return -1;
	return 0;
}


#define LEFT_CHILD(i)   (((i) * 2) + 1) 
int
prioq_heapify(struct prioq *queue)
{
	unsigned int index, largest;	
	struct prioq_head*  tmp;
	
	for (index = 0,largest = LEFT_CHILD(index);
			largest < queue->used;
			index = largest, largest = LEFT_CHILD(index)) {
		
		if (queue->compare(queue->heap[largest],
			queue->heap[largest+1]) < 0) {
            
			largest++;			
		}
		if (queue->
			compare(queue->heap[index],
			queue->heap[largest])<0) {
			
			tmp = queue->heap[largest];
			queue->heap[largest] = queue->heap[index];
			queue->heap[largest]->index = largest; 
			queue->heap[index]          = tmp;
			queue->heap[index]->index   = index; 
		}else 
			return 0;
	}	
	if ((largest == queue->used)
		&& (queue->compare(queue->heap[index],
		queue->heap[largest]) < 0)) {
        
		tmp = queue->heap[largest];
		queue->heap[largest]        = queue->heap[index];
		queue->heap[largest]->index = largest; 
		queue->heap[index]          = tmp;
		queue->heap[index]->index   = index; 
	}
	return 1;	
}



void
prioq_remove(struct prioq *queue, struct prioq_head *hp) 
{
    
	if (queue == NULL || hp == NULL) {
		fprintf(stderr, "prioq_remove: NULL args\n");
	}
	
	if (hp->index > queue->used)
		fprintf(stderr,"prioq_remove: index out of range\n");
	
	int index = hp->index;
		
	while (index > 0) {
		queue->heap[index]        = queue->heap[PARENT(index)];
		queue->heap[index]->index = PARENT(index); 
		index = PARENT(index);
	}
	queue->used--;
	int used       = queue->used;
	queue->heap[0] = queue->heap[used];
	queue->heap[0]->index = 0;
	prioq_heapify(queue);
	//queue->heap[used] = NULL;
	
}	

void
prioq_free(struct prioq** queue)
{
	if (*queue == NULL) {
		fprintf(stderr,"prioq_free: The queue is free\n");
		return;
	}
	if ((*queue)->heap == NULL) {
		free(*queue);
		fprintf(stderr,"prioq_free: The queue id free\n");
		return;
	}
	
	free((*queue)->heap);
	free(*queue);
	*queue = NULL;
	fprintf(stderr, "prioq_free: The queue is free\n");
}	
	
	
   
	

struct prioq_head *
prioq_peek(struct prioq* queue) 
{
	struct prioq_head *head;
	struct user_data  *ptr;
	
	if (queue == NULL) {
		fprintf(stderr,"prioq_peek: The queue is NULL\n");
		return NULL;
	}		
	if (queue->used == 0) 
		return NULL;		
	
	head        = queue->heap[0];
	head->index = 0;
	ptr         = (struct user_data*)head;
	fprintf(stderr,"prioq_peek: data %d index %d\n",
			             ptr->data,ptr->head.index);
	return head;
}

struct prioq_head *
prioq_get(struct prioq *queue)
{
	struct prioq_head *head;
	struct user_data  *ptr;
	unsigned int used;
	    
	if (queue == NULL) {
		fprintf(stderr,"prioq_get: Can't extract head from queue\n");
		return NULL;
	}	
	if (queue->used == 0)
		return NULL;		
 	
	head = queue->heap[0];
	head->index = 0;
	ptr  = (struct user_data*)head;
	used = queue->used - 1;
	
	fprintf(stderr,"prioq_get: data %d\n",ptr->data);
	
	queue->heap[0]        = queue->heap[used];
	queue->heap[0]->index = 0;
	queue->used--;
	prioq_heapify(queue);
	//queue->heap[used] = NULL;	
	return head;	
}	

int main(void){
	struct prioq *q = prioq_new(compare); 
	struct user_data a,b,c,d,e,f;
	
	unsigned int i;		
	a.data = 0;
	b.data = 1;
	c.data = 2;
	d.data = 3;
	e.data = 4;
	f.data = 5;
    
	prioq_put(q, (struct prioq_head*) &a);
	prioq_put(q, (struct prioq_head*) &b);
	prioq_put(q, (struct prioq_head*) &c);
	prioq_get(q);

	
	for (i = 0; i < q->used; i++) {
		printf("index = %d data = %d\n",(q->heap[i])->index,
				((struct user_data *)q->heap[i])->data);
	}
	
	return 0;
}
  	
